-- see https://markkarpov.com/tutorial/megaparsec.html
{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE OverloadedStrings #-}

module Main.Parser.Megaparsec (parse) where

import Control.Monad.Combinators.Expr
import Data.Functor.Identity qualified
import Data.Text
import Data.Void (Void)
import Main.Types.Expr
import Text.Megaparsec as MP hiding (parse)
import Text.Megaparsec qualified as MP
import Text.Megaparsec.Char qualified as C
import Text.Megaparsec.Char.Lexer qualified as L

type Parser = Parsec Void Text

lexeme :: Parser a -> Parser a
lexeme = L.lexeme C.space

symbol :: Text -> Parser Text
symbol = L.symbol C.space

int :: Parser Int
int = lexeme $ L.signed (return ()) L.decimal

term :: Parser Expr
term = Lit <$> int

table :: [[Operator Parser Expr]]
table =
  [ [ binaryOp "*" Mul,
      binaryOp "/" Div
    ],
    [ binaryOp "+" Add,
      binaryOp "-" Sub
    ]
  ]

binaryOp :: Text -> (a -> a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
binaryOp name f = InfixL (f <$ symbol name)

prefixOp :: Text -> (a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
prefixOp name f = Prefix (f <$ symbol name)

postfixOp :: Text -> (a -> a) -> Operator (ParsecT Void Text Data.Functor.Identity.Identity) a
postfixOp name f = Postfix (f <$ symbol name)

expr :: Parser Expr
expr = makeExprParser term table

parseExpr :: Text -> Either (ParseErrorBundle Text Void) Expr
parseExpr = MP.parse (C.space *> expr <* eof) ""

parse :: Text -> Expr
parse t = do
  case parseExpr t of
    Right r -> r

-- TODO: add error handling